import os
import json
import datetime
import requests
import warnings

import pandas as pd

ACCESS_TOKEN_ENDPOINT = "https://www.reddit.com/api/v1/access_token"
BASE_ENDPOINT = "https://oauth.reddit.com"

USER_AGENT = "LSE DS105W API practice by {username}"

def authenticate(credentials_json_path: str) -> str:
    """Authenticate with Reddit using credentials stored in a JSON file.

    Args:
        credentials_json_path: Path to a JSON file containing the following keys:
            - reddit_username
            - reddit_password
            - app_client_id
            - app_client_secret

    Returns:
        A dictionary containing the necessary headers to authenticate with Reddit.
        This dictionary contains the following keys:
            - User-Agent
            - Authorization
    """

    # open the file and load the data into a variable
    with open(credentials_json_path, "r") as f:
        credentials = json.load(f)

    client_auth = requests.auth.HTTPBasicAuth(credentials["app_client_id"], 
                                              credentials["app_client_secret"])

    # You also need to send, via HTTP POST, your Reddit username and password.
    post_data = {"grant_type": "password", 
                 "username": credentials["reddit_username"],
                 "password": credentials["reddit_password"]}

    headers = {"User-Agent": USER_AGENT.format(username=credentials["reddit_username"])}

    response = requests.post(ACCESS_TOKEN_ENDPOINT, auth=client_auth, data=post_data, headers=headers)
    reddit_token = response.json()['access_token']

    headers["Authorization"] = f"bearer {reddit_token}"

    return headers

def normalise_page(page):
    first_round = pd.DataFrame.from_dict(page['data']['children'])
    final_df = pd.json_normalize(first_round['data'], max_level=1)

    date_cols = [col for col in final_df.columns if 'utc' in col]
    for col in date_cols:
        final_df[col] = final_df[col].apply(lambda x: None if x is None else datetime.datetime.fromtimestamp(x))

    return final_df


def request_listing(endpoint: str, headers: dict, params: dict = None) -> dict:
    """Make a request to the Reddit API to a listing endpoint.

    The Reddit API has this concept of "listing" endpoints, endpoints that return a list of items.

    Read more about it here: https://www.reddit.com/dev/api/#listings

    Args:
        endpoint: The endpoint to request.
        headers: A dictionary containing the necessary headers to authenticate with Reddit.
        params: A dictionary containing the parameters to send with the request.

    Returns:
        A dictionary containing the JSON response from the Reddit API.
    """

    endpoint_full_url = os.path.join(BASE_ENDPOINT, endpoint)
    response = requests.get(endpoint_full_url, headers=headers, params=params)
    page = response.json()

    after = page["data"]["after"]
    page = normalise_page(page)

    return page, after


def get_top_posts(t: str, n_posts: int, headers: dict) -> pd.DataFrame:
    """Get the top posts (right now) from Reddit's front page.

    Using the endpoint https://www.reddit.com/dev/api/#GET_top

    Args:
        token: A Reddit API access token.
    
    Returns:
        A pandas DataFrame containing the top posts.
    """
    
    if n_posts <= 0 or n_posts is None:
        warnings.warn("n_posts must be a positive integer. Defaulting to 100.")
        n_posts = 100

    # Divide the number of posts by 100, since the maximum number of posts per request is 100
    n_requests = n_posts // 100 + 1
    after = None

    # There is no scaping of a for loop here,
    # since we need the 'after' parameter from the previous request 
    # to make the next one
    results = []
    for i in range(n_requests):
        params = {"limit": 100, "t": t, "after": after}
        page, after = request_listing("top", headers, params)
        results.append(page)

    results = pd.concat(results, ignore_index=True)
    return results.iloc[:n_posts]




    
